package be.jeffcheasey88.peeratcode.parser.state.generator;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.Tokenizer;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.Element.SynthaxeComposant;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.Element.SynthaxeComposant.Composant.ComposantCall;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.Element.SynthaxeComposant.Composant.NamedElement;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.Element.SynthaxeComposant.Composant.Redirection;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.Element.SynthaxeComposant.Composant.Validator;

public class ParserComposantGenerator {
	
	public static void main(String[] args) throws Exception{
		ParserComposantGenerator generator = new ParserComposantGenerator(new BufferedReader(new FileReader(new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-parseur\\java.pc"))));
		generator.generate();
	}
	
	private BufferedReader reader;
	
	public ParserComposantGenerator(BufferedReader reader){
		this.reader = reader;
	}
	
	public void generate() throws Exception{
		SynthaxeComposant synthax = load();
	}
	
	private SynthaxeComposant load() throws Exception{
		SynthaxeComposant composant = new SynthaxeComposant();
		Parser<Element> parser = new Parser<Element>(){
			{
				StateTree<Element> validator = new StateTree<>();
				StateTree<Element> modifiers = new StateTree<>();
				modifiers.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				StateTree<Element> end = modifiers.then((v) -> v.validate(
						(t) -> t.getValue().equals("="),
						(bag,t) -> bag.set("end", true)));
				StateTree<Element> loop = modifiers.then((v) -> v.validate(
						(t) -> t.getValue().equals("*"),
						(bag,t) -> bag.set("loop", true)));
				loop.then(end);
				loop.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				end.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				
				StateTree<Element> simpleAction = validator
					.then((v) -> v.validate((t) -> t.getValue().equals("[")))
					.then((v) -> v.validate((t) -> !t.getValue().equals("]"), (bag, t) -> bag.set(t))).loop()
					.then((v) -> v.validate((t) -> t.getValue().equals("]"), (bag, t) -> bag.set("type", "simple")));
				StateTree<Element> customAction = validator
						.then((v) -> v.validate((t) -> t.getValue().equals("{")))
						.then((v) -> v.validate((t) -> !t.getValue().equals("}"), (bag, t) -> bag.set(t))).loop()
						.then((v) -> v.validate((t) -> t.getValue().equals("}"), (bag, t) -> bag.set("type", "custom")));
				simpleAction.then(modifiers);
				customAction.then(modifiers);
				
				StateTree<Element> namedValidator = new StateTree<>();
				StateTree<Element> leftSide = namedValidator
					.then((v) -> v.validate((t) -> t.getValue().equals("-")) && v.validate((t) -> t.getValue().equals(">")))
					.then((v) -> v.validate(
							(t) -> t.getType().equals(TokenType.NAME),
							(bag, t) -> bag.set(t)));
				leftSide.end((composant, bag) -> {
					Element result = new NamedElement(bag);
					bag.set(result);
					return result;
				});
				leftSide.then((v) -> 
						v.validate((t) -> t.getValue().equals("(")) &&
						v.validate((t) -> t.getType().equals(TokenType.NAME), (bag, t) -> bag.set("inside", t)) &&
						v.validate((t) -> t.getValue().equals(")"))
						)
						.end((composant, bag) -> {
							Element result = new NamedElement(bag);
							bag.set(result);
							return result;
						});
				
				StateTree<Element> gotoValidator = new StateTree<>();
				gotoValidator.then((v) -> v.validate(
						(t) -> t.getType().equals(TokenType.NAME),
						(bag, t) -> bag.set(t)))
					.then((v) -> v.validate((t) -> t.getValue().equals(":")))
					.end((composant, bag) -> composant);
				
				StateTree<Element> redirection = new StateTree<>();
				redirection.then((v) -> v.validate((t) -> 
					 t.getType().equals(TokenType.NAME),
					 (bag, t) -> bag.set(t)
				)).end((composant, bag) -> {
					Element result = new Redirection(bag);
					bag.set(result);
					return result;
				});
				
				StateTree<Element> methodCall = new StateTree<>();
				methodCall.then((v) -> v.validate((t) -> t.getValue().equals(":")))
						.then((v) -> v.validate(
								(t) -> t.getType().equals(TokenType.NAME),
								(bag, t) -> bag.set(t)))
						.end((composant, bag) -> {
							Element result = new ComposantCall(bag);
							bag.set(result);
							return result;
						});
				
				StateTree<Element> operations = new StateTree<>();
				StateTree<Element> operation = new StateTree<>();
				
				StateTree<Element> op_childs = new StateTree<>();
				op_childs.then((v) -> v.validate((t) -> t.getValue().equals("(")))
					.then(new RedirectStateTree<>(operations, (bag) -> {
						return "operations";
					}))
					.then((v) -> v.validate((t) -> t.getValue().equals(")")))
					.end((composant, bag) -> {
						System.out.println("\t"+bag);
						return composant;
					});
				
				operation.then(new RedirectStateTree<>(gotoValidator, (bag) -> "?"))
				.then(operations);
				
				StateTree<Element> op_validator = operation.then(new RedirectStateTree<>(validator, (bag) -> "validator"));
				op_validator.end((composant, bag) -> {
					return bag.<Bag>get("validator").<Element>get();
				}).then(op_childs);
				op_validator.then(new RedirectStateTree<>(namedValidator, (bag) -> "namedValidator")).end((composant, bag) -> {
					bag.<Bag>get("namedValidator").<NamedElement>get().child = bag.<Bag>get("validator").<Validator>get();
					return composant;
				}).then(op_childs);
				
				StateTree<Element> op_call = operation.then(new RedirectStateTree<>(methodCall, (bag) -> "methodCall"));
				op_call.end((composant, bag) -> {
					return bag.<Bag>get("methodCall").<Element>get();
				}).then(op_childs);
				op_call.then(new RedirectStateTree<>(namedValidator, (bag) -> "namedValidator")).end((composant, bag) -> {
					bag.<Bag>get("namedValidator").<NamedElement>get().child = bag.<Bag>get("methodCall").<Validator>get();
					return composant;
				}).then(op_childs);
				
				StateTree<Element> op_redirect = operation.then(new RedirectStateTree<>(redirection, (bag) -> "redirection"));
				op_redirect.end((composant, bag) -> {
					return bag.<Bag>get("redirection").<Element>get();
				}).then(op_childs);
				op_redirect.then(new RedirectStateTree<>(namedValidator, (bag) -> "namedValidator")).end((composant, bag) -> {
					bag.<Bag>get("namedValidator").<NamedElement>get().child = bag.<Bag>get("redirection").<Validator>get();
					return composant;
				}).then(op_childs);
				
				StateTree<Element> suit = operations.then(new RedirectStateTree<>(operation, (bag) -> "operation")).loop();
				suit.then((v) -> v.validate((t) -> t.getValue().equals(",")))
					.then(suit).end((composant, bag) -> composant);
				suit.then((v) -> v.validate((t) -> t.getValue().equals(";")))
					.then(suit).end((composant, bag) -> composant);
				suit.end((composant, bag) -> composant);
				
				StateTree<Element> main = new StateTree<>();
				
				StateTree<Element> composant = main.then((v) -> v.validate(
						(token) -> token.getType().equals(TokenType.NAME),
						(bag, token) -> bag.set(token)
				));
				composant.then(new RedirectStateTree<>(op_childs, (bag) -> "content"))
				.end((element, bag) -> {
					System.out.println(bag);
					return element;
				})
				.then(composant);
				
				
				setTokenizer(new Tokenizer());
				setStateTree(main);
			}
		};
		
		parser.parse(reader, composant);
		parser.build(new BufferedWriter(new FileWriter(new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-parseur\\java.out"))));
		
		return composant;
	}
	
	public static class Element{
		
		public Element(){
			
	}
	
	public static class SynthaxeComposant extends Element{
		
		private Map<String, Composant> composants;
		
		public SynthaxeComposant(){
			this.composants = new HashMap<>();
		}
		
		void register(String name, Composant composant){
			this.composants.put(name, composant);
		}
		
		public static class Composant extends Element{
			
			private String title;
			
			private List<Element> path;
			private List<Element> declared;
			private List<Element> next;
			
			public Composant(){
				
			}
			
				public static class Validator extends Element{
					
					private String action;
					private boolean end;
					private boolean loop;
					
					public Validator(Bag bag){
						String type = bag.get("type");
						action = bag.<Token>get().getValue();
						if(type.equals("simple")){
							action = '['+action+']';
						}else{
							action = '{'+action+'}';
						}
						this.end = (bag.get("end") != null);
						this.loop = (bag.get("loop") != null);
						System.out.println(this);
					}
					
					@Override
					public String toString(){
						return action+((loop ? "*":""))+((end ? "=":""));
					}
				}
				
				public static class NamedElement extends Element{
					
					private String logic;
					private String name;
					
					private Element child;
					
					public NamedElement(Bag bag){
						Token inside = bag.get("inside");
						if(inside != null){
							this.logic = bag.<Token>get().getValue();
							this.name = inside.getValue();
						}else{
							this.name = bag.<Token>get().getValue();
						}
					}
					
					@Override
					public String toString(){
						return child+" -> "+((logic == null) ? name : logic+"("+name+")");
					}
					
				}
				
				public static class Redirection extends Element{
					
					private String target;
					
					public Redirection(Bag bag){
						this.target = bag.<Token>get().getValue();
						System.out.println(this);
					}
					
					@Override
					public String toString(){
						return target;
					}
				}
				
				public static class ComposantCall extends Element{
					
					private String name;
					
					public ComposantCall(Bag bag){
						this.name = bag.<Token>get().getValue();
						System.out.println(this);
					}
					
					@Override
					public String toString(){
						return ":"+name;
					}
				}
				
			}
		}
	}
}
